/*
 * Copyright (c)1997 by Dianne Hackborn.
 * All rights reserved.
 *
 * Various classes and definitions that I find useful.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Dianne Hackborn,
 * at <hackbod@lucent.com> or <hackbod@enteract.com>.
 *
 * ----------------------------------------------------------------------
 *
 * ArpMessage.h
 *
 * A slightly more convenient interface to constructing and
 * accessing BMessage objects.  This includes two things:
 *
 * New methods for adding and retrieving BFont and rgb_color
 * structures from a BMessage.
 *
 * A simpler set of functions for placing data in a BMessage,
 * useful to avoid creating temporary variables.  It basically
 * works like this:
 *
 * FunctionThatWantsMessage( ArpMessage()
 *                           .SetInt32(234)
 *                           .SetString("A string")
 *                           ... );
 *
 * ----------------------------------------------------------------------
 *
 * Known Bugs
 * ~~~~~~~~~~
 *
 *	• ARP_FONT_TYPE should store font and style names rather
 *	  than their ID codes.
 *
 * ----------------------------------------------------------------------
 *
 * To Do
 * ~~~~~
 *
 * ----------------------------------------------------------------------
 *
 * History
 * ~~~~~~~
 *
 * 0.1: Created this file.
 *
 */

#pragma once

#ifndef ARPCOMMON_ARPMESSAGE_H
#define ARPCOMMON_ARPMESSAGE_H

#ifndef _MESSAGE_H
#include <Message.h>
#endif

#ifndef _FONT_H
#include <Font.h>
#endif

enum {
	ARP_FONT_TYPE		= 'font',
};

class ArpMessage : public BMessage {

  public:

	ArpMessage();
	ArpMessage(uint32 what);
	ArpMessage(const BMessage& a_message);
	ArpMessage(BMessage *a_message);
	
	ArpMessage&	operator=(const BMessage& msg);

	ArpMessage& Update(const BMessage& msg);
		
	// Standard types there is no BMessage interface for
	status_t	AddRGBColor(const char *name, const rgb_color *col);
	status_t	FindRGBColor(const char *name, rgb_color *col) const;
	status_t	FindRGBColor(const char *name, int32 index, rgb_color *col) const;
	
	// My custom types
	status_t	AddFont(const char *name, const BFont *font);
	status_t	FindFont(const char *name, BFont *font) const;
	status_t	FindFont(const char *name, int32 index, BFont *font) const;
	
	ArpMessage&	SetRect(const char *name, const BRect& a_rect);
	ArpMessage&	SetPoint(const char *name, const BPoint& a_point);
	ArpMessage&	SetString(const char *name, const char* a_string);
	ArpMessage&	SetInt8(const char *name, int8 val);
	ArpMessage&	SetInt16(const char *name, int16 val);
	ArpMessage&	SetInt32(const char *name, int32 val);
	ArpMessage&	SetInt64(const char *name, int64 val);
	ArpMessage&	SetBool(const char *name, bool a_boolean);
	ArpMessage&	SetFloat(const char *name, float a_float);
	ArpMessage&	SetDouble(const char *name, double a_double);
	ArpMessage&	SetPointer(const char *name, const void *ptr);
	ArpMessage&	SetMessenger(const char *name, const BMessenger& messenger);
	ArpMessage&	SetRef(const char *name, const entry_ref *ref);
	ArpMessage&	SetMessage(const char *name, const BMessage *msg);
	ArpMessage&	SetRGBColor(const char *name, const rgb_color *col);
	ArpMessage&	SetFont(const char *name, const BFont *font);
	ArpMessage&	SetData(const char *name,
						type_code type,
						const void *data,
						ssize_t numBytes);

	const BRect&		GetRect(const char *name, const BRect& def, int32 index=0) const;
	const BPoint&		GetPoint(const char *name, const BPoint& def, int32 index=0) const;
	const char*			GetString(const char *name, const char* def, int32 index=0) const;
	int8				GetInt8(const char *name, int8 def, int32 index=0) const;
	int16				GetInt16(const char *name, int16 def, int32 index=0) const;
	int32				GetInt32(const char *name, int32 def, int32 index=0) const;
	int64				GetInt64(const char *name, int64 def, int32 index=0) const;
	bool				GetBool(const char *name, bool def, int32 index=0) const;
	float				GetFloat(const char *name, float def, int32 index=0) const;
	double				GetDouble(const char *name, double def, int32 index=0) const;
	void*				GetPointer(const char *name, void* def, int32 index=0) const;
	const BMessenger&	GetMessenger(const char *name, const BMessenger& def, int32 index=0) const;
	const entry_ref&	GetRef(const char *name, const entry_ref& def, int32 index=0) const;
	const BMessage&		GetMessage(const char *name, const BMessage& def, int32 index=0) const;
	const BFont&		GetFont(const char *name, const BFont& def, int32 index=0) const;
	const rgb_color&	GetRGBColor(const char *name, const rgb_color& def, int32 index=0) const;
	const void*			GetData(const char *name, type_code type,
								ssize_t* gotBytes, const void* def,
								int32 index=0) const;

	status_t GetError(void) { return status; }
	void ClrError(void) { status = B_NO_ERROR; }
	
  private:
	status_t status;
};

/* ------------------------------------------------------------- */

inline ArpMessage::ArpMessage()
	: BMessage(), status(B_NO_ERROR)
{ }

inline ArpMessage::ArpMessage(uint32 what)
	: BMessage(what), status(B_NO_ERROR)
{ }

inline ArpMessage::ArpMessage(const BMessage& msg)
	: BMessage(msg), status(B_NO_ERROR)
{ }

inline ArpMessage::ArpMessage(BMessage* msg)
	: BMessage(msg), status(B_NO_ERROR)
{ }

inline ArpMessage&	ArpMessage::operator=(const BMessage& msg)
{
	*(BMessage*)this = msg;
	status = B_NO_ERROR;
	return *this;
}

#endif
