/*
 * Copyright (c)1997 by Dianne Hackborn.
 * All rights reserved.
 *
 * Based (increasingly loosly) on WebTerm, a Web-based terminal
 * applet written in Java, which is
 * Copyright (C)1996 by the National Alliance for Computational
 * Science and Engineering (NACSE).
 * See <URL:http://www.nacse.org/> for more information.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Dianne Hackborn,
 * at <hackbod@lucent.com> or <hackbod@enteract.com>.
 *
 * ----------------------------------------------------------------------
 *
 * ArpTerminal.h
 *
 * The ArpTerminal class adds all of the functionality to
 * ArpCoreTerminal that is needed to make a fully usable
 * terminal widget.  In particular, it allows the user to
 * highlight regions with the mouse and perform cut&copy
 * operations, implements an interface to ArpEmulator, and
 * transmits user input to the connected emulator.
 *
 * ----------------------------------------------------------------------
 *
 * Known Bugs
 * ~~~~~~~~~~
 *
 *	• There is code in the class to extract character styles
 *	  along with the text when doing a cut...  however, the
 *	  data we create seems to crash StyledEdit so this feature
 *	  is currently disabled.
 *
 * ----------------------------------------------------------------------
 *
 * To Do
 * ~~~~~~~~~~
 *
 *	• The paste prompt should be placed as close as possible
 *	  to the current cursor position.
 *
 * ----------------------------------------------------------------------
 *
 * History
 * ~~~~~~~
 *
 * 8/16/1997:
 *	• Now positions the paste query display at the current
 *	  cursor location, so that it is hopefully more obvious...
 *	  Now includes style information with copied text, which
 *	  StyledEdit can pick up.  Yay!
 *	• The initial highlight position is now moved during a screen
 *	  scroll.
 *
 * 7/25/1997:
 *	• Was transforming a LF into CR LF, and shouldn't have been.
 *	• Accidently left some debugging output in Copy/Paste.
 *	• SelectAll() was not selecting all.
 *
 * 7/17/1997:
 *	• Split basic terminal functionality off into
 *	  ArpCoreTerminal; this is basically all new code since then.
 *
 */

#pragma once

#ifndef ARPCOMMON_ARPTERMINAL_H
#define ARPCOMMON_ARPTERMINAL_H

#ifndef _LIST_H
#include <BList.h>
#endif

#ifndef _MENU_H
#include <Menu.h>
#endif

#ifndef ARPCOMMON_ARPMOUSETRACKER_H
#include <ArpCommon/ArpMouseTracker.h>
#endif

#ifndef ARPCOMMON_ARPCORETERMINAL_H
#include <ArpCommon/ArpCoreTerminal.h>
#endif

#ifndef ARPCOMMON_ARPEMULATOR_H
#include <ArpCommon/ArpEmulator.h>
#endif

class ArpTerminal :	public ArpCoreTerminal,
					protected ArpMouseTracker {

public:

	ArpTerminal(BRect frame, const char* name,
			 uint32 resizeMode = B_FOLLOW_ALL_SIDES,
			 uint32 flags = B_WILL_DRAW|B_FRAME_EVENTS|B_NAVIGABLE);
	virtual ~ArpTerminal();
	
	/* Set the emulator associated with this terminal.
	   Once passed in, the terminal owns the emulator object
	   until another one is supplied -- any attached emulator
	   will be deleted if the terminal is deleted.
	   
	   SetEmulator() returns the emulator that was last attached
	   to the terminal -- it no longer belongs to the terminal,
	   so you usually want to delete the last emulator thusly:
	   "delete myTerminal.SetEmulator(myNewEmulator);"
	   A NULL can be passed in to selecte the default base
	   ArpEmulator class; a NULL will then be returned when
	   SetTerminal is called next.
	*/
	ArpEmulatorInterface* SetEmulator(ArpEmulatorInterface* emulator);
	
	/* Get the emulator associated with this terminal.
	   Guaranteed to always return a valid value.
	*/
	ArpEmulatorInterface& Emulator(void);

	bool IsDefaultEmulator(void);
	
	/* These functions are used to turn paste verification on
	   and off.  When on, any time a paste operation occurs
	   a view is popped up within the terminal that allows the
	   user to see what text is in the clipboard and verify
	   that the operation should proceed.  The hope is that this
	   will be useful when using the terminal over slow remote
	   connections, since accidently pasting a large clipboard
	   of text can be extremely frustrating and an undo
	   capability is usually not provided.
	*/
	void SetPasteVerified(bool state);
	bool PasteVerified(void);
	
	/* This is a convenience function to update the state
	   of a menu containing the available emulation names.
	   It turns the check mark on the menu with the same
	   name as the current emulation, and turns all others
	   off.  (Well, it's a little more complicated than that,
	   but that's all I'll explain for now... ;)
	*/
	bool CheckEmulationMenu(BMenu* inmenu);
	
  /* ------------------------------------------------------------
     ARPTERMINALINTERFACE METHODS WE OVERRIDE.
     ------------------------------------------------------------ */

	// Need to track scrolling when user is highlighting text.
	void TermScrollRegion(int32 top, int32 bottom, int32 amount);
										
	virtual void TermSetHighlight(TermHighlightType type,
									int32 row1, int32 col1,
									int32 row2, int32 col2);

	/* By default, we have these echo text right back out to
	   the emulator.  You will almost definitely want to override
	   this. */
	virtual void TermSendRemote(const ichar * d, size_t len);

  /* ------------------------------------------------------------
     BVIEW METHODS WE OVERRIDE.
     ------------------------------------------------------------ */

	virtual void MessageReceived(BMessage* message);
	
	virtual void TermSizeChanged(int32 rows, int32 cols);
	virtual void TermEmulationChanged(const char* name);
	virtual	void ScrollTo(BPoint where);
	
	virtual void KeyDown(const char *bytes, int32 numBytes);
	virtual void MouseDown(BPoint point);
	virtual void TrackMouseMoved(BPoint where, uint32 buttons);
	virtual void TrackMousePulse(BPoint where, uint32 buttons);
	virtual void TrackMouseUp(BPoint where, uint32 buttons);
	
	virtual void MakeFocus(bool focusState = TRUE);

	// A clipboard interface like BTextView
	virtual void Cut(BClipboard *clipboard);
	virtual void Copy(BClipboard *clipboard);
	virtual void Paste(BClipboard *clipboard);
	virtual void SelectAll(void);
	
	/* ------------------------------------------------------------
	   INHERITANCE SECTION
	   ------------------------------------------------------------ */

protected:

	void ShowPasteReq(const char* text, ssize_t size);
	
	/* ------------------------------------------------------------
	   IMPLEMENTATION SECTION
	   ------------------------------------------------------------ */

private:
	
	void SetPastePos(void);
	void DoPaste(const char* text, size_t size);
	void DidInput(void);
	
	/* ------------------------------------------------------------
	   INSTANCE DATA SECTION
	   ------------------------------------------------------------ */

	//BList addons;
	
	ArpEmulatorInterface* curEmulator;
	ArpEmulator defEmulator;

	bool pasteVerify;
		
	bool wasNavigable;
	BView* pasteView;
	BTextView* pasteText;
	BScrollView* pasteScroll;
	BButton* cancelButton;
	BButton* pasteButton;
	
	int32 selHMode;				// Current highlight mode / off
	BPoint hInitPoint;			// Initial place button was pressed
	int32 hInitRow, hInitCol;	// Cursor location corresponding to above
	bool selHShown;				// Set if highlight is occurring.
	
	// This is just for debugging
#ifdef DEBUG
	static const bool outputecho;
#endif
};

#endif
