/*
 * Copyright (c)1997 by Dianne Hackborn.
 * All rights reserved.
 *
 * Based (increasingly loosly) on WebTerm, a Web-based terminal
 * applet written in Java, which is
 * Copyright (C)1996 by the National Alliance for Computational
 * Science and Engineering (NACSE).
 * See <URL:http://www.nacse.org/> for more information.
 *
 * This code is not public domain, nor freely distributable.
 * Please direct any questions or requests to Dianne Hackborn,
 * at <hackbod@lucent.com> or <hackbod@enteract.com>.
 *
 * ----------------------------------------------------------------------
 *
 * ArpTerminalInterface.h
 *
 * The ArpTerminalInterface class describes an abstract interface
 * to some terminal-like device.  It is the interface through
 * which particular emulator implementations may manipulate
 * the terminal they are connected to: it defines various
 * functions for manipulating the terminal screen and
 * getting/setting general state information.
 *
 * ----------------------------------------------------------------------
 *
 * Known Bugs
 * ~~~~~~~~~~
 *
 * ----------------------------------------------------------------------
 *
 * History
 * ~~~~~~~
 *
 * 0.1: Created this file from the WebTerm source.
 *
 */

#pragma once

#ifndef ARPCOMMON_ARPTERMINALINTERFACE_H
#define ARPCOMMON_ARPTERMINALINTERFACE_H

// This is only to get the ichar definition
#ifndef ARPCOMMON_ARPSTRING_H
#include <ArpCommon/ArpString.h>
#endif

class ArpTerminalInterface {

public:

	/* Mode flags.  These are global operating modes of the terminal;
	   the initial state of all modes is off. */

	enum {
		TERM_MODESMOOTH = 1<<0,    // Smooth scroll display?
		TERM_MODENEWLINE = 1<<1,   // Go to left column on NL?
		TERM_MODENOWRAP = 1<<2,    // Wrap at right of screen?
		TERM_MODEINVERSE = 1<<3,   // Inverse video display?
		TERM_MODEHIDECURSOR = 1<<4,// Make the cursor invisible?
	};
	typedef uint32 TermModeType;
	
	/* Output flags used with the TermSendTTY() methods. */

	enum {
		TERM_OUTRAW     = 1<<0,  // Don't interpret \n \b etc.?
		TERM_OUTPARTIAL = 1<<1,  // Don't update screen?
	};
	typedef uint32 TermOutputType;
	
	/* Style flags.  Use with TermSetStyle() and TermStyle(); each character can
	   have its own style, and that style is taken from the current style
	   setting when it was written to the terminal. */

	enum {
		TERM_STYLEPLAIN      = 0,
		TERM_STYLEINVERSE    = 1<<0,
		TERM_STYLEBOLD       = 1<<1,
		TERM_STYLEITALIC     = 1<<2,
		TERM_STYLEUNDERSCORE = 1<<3,
		TERM_STYLEMASK       = 0xff,

		TERM_STYLECOLOR_MASK = 0xf,
		TERM_STYLEFOREGROUND_POS = 8,
		TERM_STYLEBACKGROUND_POS = 12,
		
		// A special flag, identifying whatever the current
		// style us.  This is used as a default value for
		// later functions.
		TERM_STYLECURRENT    = 1<<7,
	};
	typedef uint16 style_t;
	
	/* Colors. */

	enum {
		// These are colors that can applied to each individual
		// character on the display.
		TERM_NORMALCOLOR = 0,
		TERM_COLOR_1 = 1,		// By default, these are ANSI colors
		TERM_COLOR_2 = 2,
		TERM_COLOR_3 = 3,
		TERM_COLOR_4 = 4,
		TERM_COLOR_5 = 5,
		TERM_COLOR_6 = 6,
		TERM_COLOR_7 = 7,
		TERM_COLOR_8 = 8,
		TERM_COLOR_A = 9,		// Extra colors, initially undefined
		TERM_COLOR_B = 10,
		TERM_COLOR_C = 11,
		TERM_COLOR_D = 12,
		TERM_COLOR_E = 13,
		TERM_COLOR_F = 14,
		TERM_COLOR_G = 15,
		
		// These are colors that apply to the entire terminal.
		TERM_HIGHLIGHTCOLOR = -1,
		TERM_CURSORCOLOR = -2,
		
		// This should probably be turned into a function that
		// queries the terminal.
		TERM_NUMTEXTCOLORS = 16,
	};
	typedef int32 TermColorID;

	/* Highlighting modes */
	
	typedef enum {
		TERM_HIGHLIGHTOFF  = 0,
		TERM_HIGHLIGHTCHAR = 1,
		TERM_HIGHLIGHTWORD = 2,
		TERM_HIGHLIGHTLINE = 3
	};
	typedef uint32 TermHighlightType;
	
	// Logical character codes
	enum {
		ANSI_NUL = 0x00,
		ANSI_SOH = 0x01,
		ANSI_STX = 0x02,
		ANSI_ETX = 0x03,
		ANSI_EOT = 0x04,
		ANSI_ENQ = 0x05,
		ANSI_ACK = 0x06,
		ANSI_BEL = 0x07,
		ANSI_BS  = 0x08,
		ANSI_HT  = 0x09,
		ANSI_LF  = 0x0a,
		ANSI_VT  = 0x0b,
		ANSI_FF  = 0x0c,
		ANSI_CR  = 0x0d,
		ANSI_SO  = 0x0e,
		ANSI_SI  = 0x0f,
		ANSI_DLE = 0x10,
		ANSI_DC1 = 0x11,
		ANSI_DC2 = 0x12,
		ANSI_DC3 = 0x13,
		ANSI_DC4 = 0x14,
		ANSI_NAK = 0x15,
		ANSI_SYN = 0x16,
		ANSI_ETB = 0x17,
		ANSI_CAN = 0x18,
		ANSI_EM  = 0x19,
		ANSI_SUB = 0x1a,
		ANSI_ESC = 0x1b,
		ANSI_FS  = 0x1c,
		ANSI_GS  = 0x1d,
		ANSI_RS  = 0x1e,
		ANSI_US  = 0x1f,
		ANSI_DEL = 0x7f,
		ANSI_IND = 0x84,
		ANSI_NEL = 0x85,
		ANSI_SSA = 0x86,
		ANSI_ESA = 0x87,
		ANSI_HTS = 0x88,
		ANSI_HTJ = 0x89,
		ANSI_VTS = 0x8a,
		ANSI_PLD = 0x8b,
		ANSI_PLU = 0x8c,
		ANSI_RI  = 0x8d,
		ANSI_SS2 = 0x8e,
		ANSI_SS3 = 0x8f,
		ANSI_DCS = 0x90,
		ANSI_PU1 = 0x91,
		ANSI_PU2 = 0x92,
		ANSI_STS = 0x93,
		ANSI_CCH = 0x94,
		ANSI_MW  = 0x95,
		ANSI_SPA = 0x96,
		ANSI_EPA = 0x97,
		ANSI_CSI = 0x9b,
		ANSI_ST  = 0x9c,
		ANSI_OSC = 0x9d,
		ANSI_PM  = 0x9e,
		ANSI_APC = 0x9f,
	};
	
	/* ------------------------------------------------------------
	   GLOBAL STATE CONTROL.
	   ------------------------------------------------------------ */

	virtual void TermReset(bool hard) = 0;

	virtual void TermSetMode(TermModeType mode) = 0;
	virtual TermModeType TermMode(void) const = 0;

	// The translation table interface needs to be reworked
	// (or simply removed) for DR9's UTF-8 stuff.
	virtual void TermSetTranslation(const ichar* t) = 0;
	virtual const ichar* TermTranslation(void) const = 0;

	virtual void TermSetTextBackground(rgb_color bg, TermColorID num) = 0;
	virtual void TermSetTextForeground(rgb_color fg, TermColorID num) = 0;
	virtual rgb_color TermTextBackground(TermColorID num) const = 0;
	virtual rgb_color TermTextForeground(TermColorID num) const = 0;

	// Fixed size is used to request a specific number of rows
	// or columns; use -1 for each to resize with view.
	virtual bool TermSetFixedSize(int32 rows, int32 cols) = 0;
	virtual void TermGetFixedSize(int32* rows, int32* cols) const = 0;
	
	// 'View' is the number of rows and columns visible on-screen.
	virtual void TermGetViewSize(int32* rows, int32* cols) const = 0;
	
	// This is the actual usuable terminal size: that is,
	// the fixed dimensions if they are set, else the view dimens.
	virtual void TermGetSize(int32* rows, int32* cols) const = 0;
	
	// Retrieve the number of rows currently in history.
	// NOTE: This is expressed in terminal coordinates, that is,
	// starting at -1 for one line, -2 for two lines, etc.
	virtual int32 TermHistoryTop(void) const = 0;
	
	/* ------------------------------------------------------------
	   SIMPLE MANIPULATION METHODS.
	   ------------------------------------------------------------ */

	virtual void TermSetCursorPos(int32 row, int32 col) = 0;
	virtual void TermGetCursorPos(int32* row, int32* col) const = 0;

	virtual int32 TermXToCol(float x) const = 0;
	virtual int32 TermYToRow(float y) const = 0;

	virtual void TermSetRegion(void) = 0;
	virtual void TermSetRegion(int32 top, int32 bottom) = 0;
	virtual void TermSetRegion(int32 top, int32 bottom, int32 left, int32 right) = 0;
	virtual void TermGetRegion(int32* top, int32* bottom, int32* left, int32* right) const = 0;

	virtual void TermSetStyle(int32 style) = 0;
	virtual void TermSetForeColor(TermColorID col) = 0;
	virtual void TermSetBackColor(TermColorID col) = 0;

	virtual int32 TermStyle(void) const = 0;
	virtual TermColorID TermForeColor(void) const = 0;
	virtual TermColorID TermBackColor(void) const = 0;

	// These two do not go through the translation table!
	virtual void TermSetChar(ichar c) = 0;
	virtual void TermSetChar(ichar c, int32 row, int32 col) = 0;

	virtual ichar TermGetChar(void) const = 0;
	virtual ichar TermGetChar(int32 row, int32 col) const = 0;
	virtual style_t TermGetCharStyle() const = 0;
	virtual style_t TermGetCharStyle(int32 row, int32 col) const = 0;

	/* ------------------------------------------------------------
	   RANGE MANIPULATION METHODS.
	   ------------------------------------------------------------ */

	/* Unlike all other methods, these allow you to access data
	   in the terminal's history: data is located at negative
	   row values, i.e. -1 is the first row of data history above
	   the actual terminal display.
	*/
	
	virtual void TermSetHighlight(TermHighlightType type,
									int32 row1, int32 col1,
									int32 row2, int32 col2) = 0;
	virtual void TermGetHighlight(TermHighlightType* type,
									int32* row1, int32* col1,
									int32* row2, int32* col2) const = 0;
	
	virtual void TermGetTextRange(int32 row1, int32 col1,
									int32 row2, int32 col2,
									size_t* array_size,
									ichar* text, style_t* style) const = 0;
									
	/* ------------------------------------------------------------
	   HIGH-LEVEL MANIPULATION METHODS.
	   ------------------------------------------------------------ */

	virtual void TermScrollRegion(int32 top, int32 bottom, int32 amount) = 0;
	virtual void TermScroll(int32 amount) = 0;

	virtual void TermSetChars(int32 left, int32 right, ichar fillc=' ',
					  style_t fillstyle=TERM_STYLECURRENT) = 0;
	virtual void TermDeleteChars(int32 left, int32 right, ichar fillc=' ',
						style_t fillstyle=TERM_STYLECURRENT) = 0;
	virtual void TermInsertChars(int32 left, int32 right, ichar fillc=' ',
						 style_t fillstyle=TERM_STYLECURRENT) = 0;

	virtual void TermClearEOL(ichar fillc=' ', style_t fillstyle=TERM_STYLECURRENT) = 0;
	virtual void TermClearBOL(ichar fillc=' ', style_t fillstyle=TERM_STYLECURRENT) = 0;
	virtual void TermClearLine(ichar fillc=' ', style_t fillstyle=TERM_STYLECURRENT) = 0;
	virtual void TermClearEOD(ichar fillc=' ', style_t fillstyle=TERM_STYLECURRENT) = 0;
	virtual void TermClearBOD(ichar fillc=' ', style_t fillstyle=TERM_STYLECURRENT) = 0;
	virtual void TermClear(ichar fillc=' ', style_t fillstyle=TERM_STYLECURRENT) = 0;

	/* ------------------------------------------------------------
	   SCREEN REFRESHING METHODS.
	   ------------------------------------------------------------ */

	virtual void TermRedraw(int32 top, int32 bot, bool full) = 0;
	virtual void TermRedraw(void) = 0;
	virtual void TermClean(void) = 0;

	/* ------------------------------------------------------------
	   CHARACTER-STREAM METHODS.
	   ------------------------------------------------------------ */

	/* These methods implement the RAW character-steam I/O
	   interface with the terminal.  They are used by emulators
	   to send raw text to the terminal display (TermSendTTY) or
	   the remote deveice it is attached to (TermSendRemote).
	   
	   If you are not using an emulator with the terminal, you
	   may call these yourself to communicate through the raw
	   data stream.  Otherwise, you probably want to use the
	   emulator functions EmulateToTTY() and EmulateToRemote(),
	   which implement the parsing and generation of appropriate
	   control characters.
	*/
	
	virtual void TermSendTTY(const ichar * d, size_t len, uint32 flags=0) = 0;
	virtual void TermSendRemote(const ichar * d, size_t len) = 0;
};

#endif
